#include "defns.h"

/**
 * Structure for the AVL Binary Search Tree Nodes.
 * Each node holds the data, the nodes' subtree height and pointers to the left and right children of the node.
 */
typedef struct _AVLTreeNode {
	Type data;						// The data
	int height;						// The subtree height
	struct _AVLTreeNode *leftChild;	// Pointer to the left child. All the nodes under this child have values <= than this nodes data.
	struct _AVLTreeNode *rightChild;	// Pointer to the right child. All the nodes under this child have values > than this nodes data.
	struct _AVLTreeNode *father;		// Pointer to the father node. If the node is the root, this is NULL.
} AVLTreeNode;

/**
 * Structure for an AVL Binary Search Tree.
 * Holds a pointer to the root node.
 */
typedef struct _AVLTree {
	AVLTreeNode *root;			// Pointer to the root node.
} AVLTree;


AVLTreeNode *createNode(Type value);
AVLTreeNode *insertNode(AVLTreeNode *root, AVLTreeNode *node);
void deleteSubtree(AVLTreeNode *root);
AVLTreeNode *findNode(AVLTreeNode *root, Type value);
void printNodeInfix(AVLTreeNode *root);
void printNodePrefix(AVLTreeNode *root);
void printNodePostfix(AVLTreeNode *root);
int subtreeHeight(AVLTreeNode *root);

AVLTreeNode *singleRotationLeft(AVLTreeNode *root);
AVLTreeNode *singleRotationRight(AVLTreeNode *root);
AVLTreeNode *doubleRotationLeft(AVLTreeNode *root);
AVLTreeNode *doubleRotationRight(AVLTreeNode *root);

AVLTree *createTree(void);
void deleteTree(AVLTree *tree);
void insertData(AVLTree *tree, Type value);
void printTree(AVLTree *tree);
AVLTreeNode *findData(AVLTree *tree, Type value);
BOOL dataExists(AVLTree *tree, Type value);
BOOL deleteData(AVLTree *tree, Type value);


